package org.opencdmp.deposit.zenodorepository.model.builder;

import gr.cite.tools.exception.MyApplicationException;
import gr.cite.tools.logging.LoggerService;
import org.opencdmp.commonmodels.enums.DmpAccessType;
import org.opencdmp.commonmodels.enums.DmpUserRole;
import org.opencdmp.commonmodels.models.DmpUserModel;
import org.opencdmp.commonmodels.models.description.*;
import org.opencdmp.commonmodels.models.descriptiotemplate.DefinitionModel;
import org.opencdmp.commonmodels.models.descriptiotemplate.fielddata.RadioBoxDataModel;
import org.opencdmp.commonmodels.models.descriptiotemplate.fielddata.SelectDataModel;
import org.opencdmp.commonmodels.models.dmp.DmpBlueprintValueModel;
import org.opencdmp.commonmodels.models.dmp.DmpModel;
import org.opencdmp.commonmodels.models.dmpblueprint.SectionModel;
import org.opencdmp.commonmodels.models.dmpreference.DmpReferenceModel;
import org.opencdmp.commonmodels.models.reference.ReferenceFieldModel;
import org.opencdmp.commonmodels.models.reference.ReferenceModel;
import org.opencdmp.deposit.zenodorepository.configuration.funder.FunderProperties;
import org.opencdmp.deposit.zenodorepository.configuration.identifier.IdentifierProperties;
import org.opencdmp.deposit.zenodorepository.configuration.pid.PidProperties;
import org.opencdmp.deposit.zenodorepository.enums.ZenodoAccessRight;
import org.opencdmp.deposit.zenodorepository.model.*;
import org.opencdmp.deposit.zenodorepository.service.zenodo.ZenodoDepositServiceImpl;
import org.opencdmp.deposit.zenodorepository.service.zenodo.ZenodoServiceProperties;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.ConfigurableBeanFactory;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import java.time.Instant;
import java.time.LocalDate;
import java.time.ZoneId;
import java.time.ZoneOffset;
import java.time.format.DateTimeFormatter;
import java.util.*;
import java.util.stream.Collectors;

@Component
@Scope(value = ConfigurableBeanFactory.SCOPE_PROTOTYPE)
public class ZenodoBuilder {
    private static final LoggerService logger = new LoggerService(LoggerFactory.getLogger(ZenodoDepositServiceImpl.class));

    private static final String UPLOAD_TYPE = "publication";
    private static final String PUBLICATION_TYPE = "datamanagementplan";
    private static final String IS_IDENTICAL_TO = "isIdenticalTo";
    private static final String CONTRIBUTOR_TYPE_RESEARCHER = "Researcher";
    private static final String CONTRIBUTOR_TYPE_PROJECT_MANAGER = "ProjectMember";
    private static final String SEMANTIC_PUBLICATION_DATE = "zenodo.publication_date";

    private final PidProperties pidProperties;
    private final IdentifierProperties identifierProperties;
    private final FunderProperties funderProperties;
    private final ZenodoServiceProperties zenodoServiceProperties;

    @Autowired
    public ZenodoBuilder(PidProperties pidProperties, IdentifierProperties identifierProperties, FunderProperties funderProperties, ZenodoServiceProperties zenodoServiceProperties){
        this.pidProperties = pidProperties;
        this.identifierProperties = identifierProperties;
        this.funderProperties = funderProperties;
	    this.zenodoServiceProperties = zenodoServiceProperties;
    }

    public ZenodoDeposit build(DmpModel dmp) {
        ZenodoDeposit deposit = new ZenodoDeposit();
        this.applyZenodoRelator(dmp, deposit);
        deposit.getMetadata().setTitle(dmp.getLabel());
        deposit.getMetadata().setUploadType(UPLOAD_TYPE);
        deposit.getMetadata().setPublicationType(PUBLICATION_TYPE);
        deposit.getMetadata().setDescription((dmp.getDescription() != null && !dmp.getDescription().isEmpty() ? dmp.getDescription() : "<p></p>"));
        deposit.getMetadata().setVersion(String.valueOf(dmp.getVersion()));
        String zenodoCommunity = zenodoServiceProperties.getCommunity();
        if(zenodoCommunity != null && !zenodoCommunity.isEmpty()) {
            if (deposit.getMetadata().getCommunities() == null) deposit.getMetadata().setCommunities(new ArrayList<>());
            ZenodoCommunity community = new ZenodoCommunity();
            community.setIdentifier(zenodoCommunity);
            deposit.getMetadata().getCommunities().add(community);
        }

        org.opencdmp.commonmodels.models.dmpblueprint.FieldModel fieldOfSemantic = this.getFieldOfSemantic(dmp, SEMANTIC_PUBLICATION_DATE);
        if (fieldOfSemantic != null){
            DmpBlueprintValueModel dmpBlueprintValueModel = this.getDmpBlueprintValue(dmp, fieldOfSemantic.getId());
            if (dmpBlueprintValueModel != null && dmpBlueprintValueModel.getDateValue() != null) {
                deposit.getMetadata().setPublicationDate(DateTimeFormatter.ofPattern("yyyy-MM-dd").withZone(ZoneId.systemDefault()).format(dmpBlueprintValueModel.getDateValue()));
            } else if (dmpBlueprintValueModel != null && dmpBlueprintValueModel.getValue() != null && !dmpBlueprintValueModel.getValue().isBlank()){
                try {
                    Instant instant = Instant.parse(dmpBlueprintValueModel.getValue());
                    deposit.getMetadata().setPublicationDate(DateTimeFormatter.ofPattern("yyyy-MM-dd").withZone(ZoneId.systemDefault()).format(instant));
                }catch (Exception e){
                    logger.error(e.getMessage(), e);
                }
            }
        }
        this.applyAccessRight(dmp, deposit);
        this.applyIsIdenticalTo(dmp, deposit);
        this.applyLicenses(dmp, deposit);
        this.applyResearchers(dmp, deposit);
        this.applyGrants(dmp, deposit);
        this.applyContributors(dmp, deposit);
        this.applyCreators(dmp, deposit);

        return deposit;
    }

    private DmpBlueprintValueModel getDmpBlueprintValue(DmpModel dmp, UUID id){
        if (dmp == null || dmp.getProperties() == null || dmp.getProperties().getDmpBlueprintValues() == null) return null;
        return dmp.getProperties().getDmpBlueprintValues().stream().filter(x-> x.getFieldId().equals(id)).findFirst().orElse(null);
    }
    
    private org.opencdmp.commonmodels.models.dmpblueprint.FieldModel getFieldOfSemantic(DmpModel dmp, String semanticKey){
        if (dmp == null || dmp.getDmpBlueprint() == null || dmp.getDmpBlueprint().getDefinition() == null || dmp.getDmpBlueprint().getDefinition().getSections() == null) return null;
        for (SectionModel sectionModel : dmp.getDmpBlueprint().getDefinition().getSections()){
            if (sectionModel.getFields() != null){
                org.opencdmp.commonmodels.models.dmpblueprint.FieldModel fieldModel = sectionModel.getFields().stream().filter(x-> x.getSemantics() != null && x.getSemantics().contains(semanticKey)).findFirst().orElse(null);
                if (fieldModel != null) return fieldModel;
            }
        }
        return null;
    }

    private List<org.opencdmp.commonmodels.models.descriptiotemplate.FieldModel> findSchematicValues(String relatedId, DefinitionModel definitionModel){
        return definitionModel.getAllField().stream().filter(x-> x.getSemantics() != null && x.getSemantics().contains(relatedId)).toList();
    }

    private List<FieldModel> findValueFieldsByIds(String fieldId, PropertyDefinitionModel definitionModel){
        List<FieldModel> models = new ArrayList<>();
        if (definitionModel == null || definitionModel.getFieldSets() == null || definitionModel.getFieldSets().isEmpty()) return models;
        for (PropertyDefinitionFieldSetModel propertyDefinitionFieldSetModel : definitionModel.getFieldSets().values()){
            if (propertyDefinitionFieldSetModel == null ||propertyDefinitionFieldSetModel.getItems() == null || propertyDefinitionFieldSetModel.getItems().isEmpty()) continue;
            for (PropertyDefinitionFieldSetItemModel propertyDefinitionFieldSetItemModel : propertyDefinitionFieldSetModel.getItems()){
                if (propertyDefinitionFieldSetItemModel == null ||propertyDefinitionFieldSetItemModel.getFields() == null || propertyDefinitionFieldSetItemModel.getFields().isEmpty()) continue;
                for (Map.Entry<String, FieldModel> entry : propertyDefinitionFieldSetItemModel.getFields().entrySet()){
                    if (entry == null || entry.getValue() == null) continue;
                    if (entry.getKey().equalsIgnoreCase(fieldId)) models.add(entry.getValue());
                }
            }
        }
        return models;
    }

    private Set<String> extractSchematicValues(List<org.opencdmp.commonmodels.models.descriptiotemplate.FieldModel> fields, PropertyDefinitionModel propertyDefinition, List<String> acceptedPidTypes) {
        Set<String> values = new HashSet<>();
        for (org.opencdmp.commonmodels.models.descriptiotemplate.FieldModel field : fields) {
            if (field.getData() == null) continue;
            List<FieldModel> valueFields = this.findValueFieldsByIds(field.getId(), propertyDefinition);
            for (FieldModel valueField : valueFields) {
                switch (field.getData().getFieldType()) {
                    case FREE_TEXT, TEXT_AREA, RICH_TEXT_AREA -> {
                        if (valueField.getTextValue() != null && !valueField.getTextValue().isBlank()) values.add(valueField.getTextValue());
                    }
                    case BOOLEAN_DECISION, CHECK_BOX -> {
                        if (valueField.getBooleanValue() != null) values.add(valueField.getBooleanValue().toString());
                    }
                    case DATE_PICKER -> {
                        if (valueField.getDateValue() != null) values.add(DateTimeFormatter.ISO_DATE.format(valueField.getDateValue()));
                    }
                    case DATASET_IDENTIFIER, VALIDATION -> {
                        if (valueField.getExternalIdentifier() != null && valueField.getExternalIdentifier().getIdentifier() != null && !valueField.getExternalIdentifier().getIdentifier().isBlank()) {
                            values.add(valueField.getExternalIdentifier().getIdentifier());
                        }
                    }
                    case TAGS -> {
                        if (valueField.getTextListValue() != null && !valueField.getTextListValue().isEmpty()) {
                            values.addAll(valueField.getTextListValue());
                        }
                    }
                    case SELECT -> {
                        if (valueField.getTextListValue() != null && !valueField.getTextListValue().isEmpty()) {
                            SelectDataModel selectDataModel = (SelectDataModel)field.getData();
                            if (selectDataModel != null && selectDataModel.getOptions() != null && !selectDataModel.getOptions().isEmpty()){
                                for (SelectDataModel.OptionModel option : selectDataModel.getOptions()){
                                    if (valueField.getTextListValue().contains(option.getValue()) || valueField.getTextListValue().contains(option.getLabel())) values.add(option.getLabel());
                                }
                            }
                        }
                    }
                    case RADIO_BOX -> {
                        if (valueField.getTextListValue() != null && !valueField.getTextListValue().isEmpty()) {
                            RadioBoxDataModel radioBoxModel = (RadioBoxDataModel)field.getData();
                            if (radioBoxModel != null && radioBoxModel.getOptions() != null && !radioBoxModel.getOptions().isEmpty()){
                                for (RadioBoxDataModel.RadioBoxOptionModel option : radioBoxModel.getOptions()){
                                    if (valueField.getTextListValue().contains(option.getValue()) || valueField.getTextListValue().contains(option.getLabel())) values.add(option.getLabel());
                                }
                            }
                        }
                    }
                    case REFERENCE_TYPES -> {
                        if (valueField.getReferences() != null && !valueField.getReferences().isEmpty()) {
                            for (ReferenceModel referenceModel : valueField.getReferences()) {
                                if (referenceModel == null
                                        || referenceModel.getType() == null || referenceModel.getType().getCode() == null || referenceModel.getType().getCode().isBlank()
                                        || referenceModel.getDefinition() == null || referenceModel.getDefinition().getFields() == null || referenceModel.getDefinition().getFields().isEmpty()) continue;
                                if (referenceModel.getType().getCode().equals(zenodoServiceProperties.getOrganizationReferenceCode()) || referenceModel.getType().getCode().equals(zenodoServiceProperties.getResearcherReferenceCode())) {
                                    if (referenceModel.getReference() != null && !referenceModel.getReference().isBlank()) {
                                        values.add(referenceModel.getReference());
                                    }
                                } else {
                                    String pid = referenceModel.getDefinition().getFields().stream().filter(x -> x.getCode() != null && x.getCode().equals(this.pidProperties.getFields().getPidName())).map(ReferenceFieldModel::getValue).findFirst().orElse(null);
                                    String pidType = referenceModel.getDefinition().getFields().stream().filter(x -> x.getCode() != null && x.getCode().equals(this.pidProperties.getFields().getPidTypeName())).map(ReferenceFieldModel::getValue).findFirst().orElse(null);
                                    if (pid != null && !pid.isBlank() && pidType != null && !pidType.isBlank() && acceptedPidTypes.contains(pidType)) {
                                        values.add(pid);
                                    }
                                }
                            }
                        }
                    }
                    case INTERNAL_ENTRIES_DESCRIPTIONS, INTERNAL_ENTRIES_DMPS, UPLOAD -> throw new MyApplicationException("Invalid type " + field.getData().getFieldType());
                    default -> throw new MyApplicationException("Invalid type " + field.getData().getFieldType());
                }
            }
        }
        return values;
    }
    
    private List<ReferenceModel> getReferenceModelOfType(DmpModel dmp, String code){
        List<ReferenceModel> response = new ArrayList<>();
        if (dmp.getReferences() == null) return response;
        for (DmpReferenceModel dmpReferenceModel : dmp.getReferences()){
            if (dmpReferenceModel.getReference() != null && dmpReferenceModel.getReference().getType() != null && dmpReferenceModel.getReference().getType().getCode() != null  && dmpReferenceModel.getReference().getType().getCode().equals(code)){
                response.add(dmpReferenceModel.getReference());
            }
        }
        return response;
    }
    
    private void applyZenodoRelator(DmpModel dmp, ZenodoDeposit deposit) {
        if (deposit.getMetadata() == null) deposit.setMetadata(new ZenodoDepositMetadata());
        
        List<String> acceptedPidTypes = this.pidProperties.getAcceptedTypes();
        List<ZenodoRelator> relatedIdentifiers = new ArrayList<>();
        for(DescriptionModel descriptionModel: dmp.getDescriptions()){
            for(String relatedId: this.identifierProperties.getRelated()){
                List<org.opencdmp.commonmodels.models.descriptiotemplate.FieldModel> fields = this.findSchematicValues(relatedId, descriptionModel.getDescriptionTemplate().getDefinition());
                Set<String> values = extractSchematicValues(fields, descriptionModel.getProperties(), acceptedPidTypes);
                for(String value: values){
                    ZenodoRelator relator = new ZenodoRelator();
                    relator.setRelation(relatedId.substring(relatedId.lastIndexOf(".") + 1));
                    relator.setIdentifier(value);
                    relatedIdentifiers.add(relator);
                }
            }
        }
        deposit.getMetadata().setRelatedIdentifiers(relatedIdentifiers);
    }
    
    private void applyAccessRight(DmpModel dmp, ZenodoDeposit deposit){
        if (deposit.getMetadata() == null) deposit.setMetadata(new ZenodoDepositMetadata());
        
        if (dmp.getAccessType() == null) {
            deposit.getMetadata().setAccessRight(ZenodoAccessRight.RESTRICTED);
            deposit.getMetadata().setAccessConditions("");
        } else {
            if (dmp.getAccessType().equals(DmpAccessType.Public)) {
                Instant publicationDate = dmp.getFinalizedAt();
                if (publicationDate == null) publicationDate = Instant.now().minusSeconds(1);

                if (publicationDate.isBefore(Instant.now())) {
                    deposit.getMetadata().setAccessRight(ZenodoAccessRight.OPEN);
                } else {
                    deposit.getMetadata().setAccessRight(ZenodoAccessRight.EMBARGOED);
                    deposit.getMetadata().setEmbargoDate(publicationDate.toString());
                }
            } else {
                deposit.getMetadata().setAccessRight(ZenodoAccessRight.RESTRICTED);
                deposit.getMetadata().setAccessConditions("");
            }
        }
    }

    private void applyIsIdenticalTo(DmpModel dmp, ZenodoDeposit deposit){
        if (deposit.getMetadata() == null) deposit.setMetadata(new ZenodoDepositMetadata());
        
        if (dmp.getAccessType().equals(DmpAccessType.Public)) {
            ZenodoRelator relator = new ZenodoRelator();
            relator.setIdentifier(zenodoServiceProperties.getDomain() + "/external/zenodo/" + dmp.getId().toString());
            relator.setRelation(IS_IDENTICAL_TO);
            if (deposit.getMetadata().getRelatedIdentifiers() == null)deposit.getMetadata().setRelatedIdentifiers(new ArrayList<>());
            deposit.getMetadata().getRelatedIdentifiers().add(relator);
        }
    }

    private void applyLicenses(DmpModel dmp, ZenodoDeposit deposit){
        if (deposit.getMetadata() == null) deposit.setMetadata(new ZenodoDepositMetadata());
       
        List<ReferenceModel> dmpLicenses = this.getReferenceModelOfType(dmp, zenodoServiceProperties.getLicensesReferenceCode());
        if (!dmpLicenses.isEmpty()) {
            for (ReferenceModel dmpLicense : dmpLicenses) {
                if (dmpLicense != null && dmpLicense.getReference() != null && !dmpLicense.getReference().isBlank()) {
                    deposit.getMetadata().setLicense(dmpLicense.getReference());
                    break;
                }
            }
        }
    }

    private void applyResearchers(DmpModel dmp, ZenodoDeposit deposit){
        if (deposit.getMetadata() == null) deposit.setMetadata(new ZenodoDepositMetadata());

        List<ZenodoContributor> researchers = new ArrayList<>();
        List<ReferenceModel> dmpResearchers = this.getReferenceModelOfType(dmp, zenodoServiceProperties.getResearcherReferenceCode());
        if (dmpResearchers != null && !dmpResearchers.isEmpty()) {
            for (ReferenceModel researcher : dmpResearchers) {
                ZenodoContributor contributor = new ZenodoContributor();
                contributor.setName(researcher.getLabel());
                contributor.setType(CONTRIBUTOR_TYPE_RESEARCHER);
                contributor.setAffiliation(researcher.getSource());
                if (researcher.getSource().equalsIgnoreCase(zenodoServiceProperties.getOrcidResearcherSourceCode())) {
                    contributor.setOrcid(researcher.getReference());
                }
                researchers.add(contributor);
            }
        }

        if (deposit.getMetadata().getContributors() == null)deposit.getMetadata().setContributors(new ArrayList<>());

        deposit.getMetadata().getContributors().addAll(researchers);
    }

    private void applyGrants(DmpModel dmp, ZenodoDeposit deposit){
        if (deposit.getMetadata() == null) deposit.setMetadata(new ZenodoDepositMetadata());
        List<ReferenceModel> dmpGrants = this.getReferenceModelOfType(dmp, zenodoServiceProperties.getGrantReferenceCode());
        List<ReferenceModel> dmpFunders = this.getReferenceModelOfType(dmp, zenodoServiceProperties.getFunderReferenceCode());

        if (!dmpGrants.isEmpty()) {
            ReferenceModel depositGrant = dmpGrants.stream().filter(x-> x.getSource().equalsIgnoreCase(zenodoServiceProperties.getOpenaireGrantSourceCode())).findFirst().orElse(null);

            if (depositGrant != null) {
                String grantReferenceTail = depositGrant.getReference().split(":")[2];
                List<FunderProperties.DoiFunder> doiFunders = this.funderProperties.getAvailable();
                if (!dmpFunders.isEmpty()) {
                    ReferenceModel depositFunder = dmpFunders.getFirst();
                    FunderProperties.DoiFunder doiFunder = doiFunders.stream()
                            .filter(doiFunder1 -> depositFunder.getLabel().contains(doiFunder1.getFunder()) || doiFunder1.getFunder().contains(depositFunder.getLabel()))
                            .findFirst().orElse(null);
                    if (doiFunder != null) {
                        String finalId = doiFunder.getDoi() + "::" + grantReferenceTail;
                        ZenodoGrant grant = new ZenodoGrant();
                        grant.setId(finalId);
                        if (deposit.getMetadata().getGrants() == null)deposit.getMetadata().setGrants(new ArrayList<>());
                        deposit.getMetadata().getGrants().add(grant);
                    }
                }
            }
        }
    }

    private void applyContributors(DmpModel dmp, ZenodoDeposit deposit){
        if (dmp.getUsers() == null) return;
        
        if (deposit.getMetadata() == null) deposit.setMetadata(new ZenodoDepositMetadata());
        
        List<ReferenceModel> dmpOrganizations = this.getReferenceModelOfType(dmp, zenodoServiceProperties.getOrganizationReferenceCode());
        String zenodoAffiliation = zenodoServiceProperties.getAffiliation();
        
        List<ZenodoContributor> contributors = new ArrayList<>();
        for (DmpUserModel userDMP: dmp.getUsers()) {
            ZenodoContributor contributor = new ZenodoContributor();
            contributor.setName(userDMP.getUser().getName());
            contributor.setType(CONTRIBUTOR_TYPE_PROJECT_MANAGER);
            if (dmpOrganizations != null && !dmpOrganizations.isEmpty()) {
                contributor.setAffiliation(dmpOrganizations.stream().map(ReferenceModel::getLabel).collect(Collectors.joining(", ")));
            } else {
                if(zenodoAffiliation != null && !zenodoAffiliation.isEmpty()) {
                    contributor.setAffiliation(zenodoAffiliation);
                }
            }
            contributors.add(contributor);
        }
        if (deposit.getMetadata().getContributors() == null)deposit.getMetadata().setContributors(new ArrayList<>());

        deposit.getMetadata().getContributors().addAll(contributors);
      
    }

    private void applyCreators(DmpModel dmp, ZenodoDeposit deposit){
        if (dmp.getUsers() == null) return;

        if (deposit.getMetadata() == null) deposit.setMetadata(new ZenodoDepositMetadata());

        List<ReferenceModel> dmpOrganizations = this.getReferenceModelOfType(dmp, zenodoServiceProperties.getOrganizationReferenceCode());
        String zenodoAffiliation = zenodoServiceProperties.getAffiliation();
        
        ZenodoCreator creator = new ZenodoCreator();
        DmpUserModel dmpModel = dmp.getUsers().stream().filter(userDMP -> userDMP.getRole().equals(DmpUserRole.Owner)).findFirst().orElse(null);
        if (dmpModel == null || dmpModel.getUser() == null) return;
        
        creator.setName(dmpModel.getUser().getName());
        if (dmpOrganizations != null && !dmpOrganizations.isEmpty()) {
            creator.setAffiliation(dmpOrganizations.stream().map(ReferenceModel::getLabel).collect(Collectors.joining(", ")));
        } else {
            if(zenodoAffiliation != null && !zenodoAffiliation.isEmpty()) {
                creator.setAffiliation(zenodoAffiliation);
            }
        }
        if (deposit.getMetadata().getCreators() == null)deposit.getMetadata().setCreators(new ArrayList<>());
        deposit.getMetadata().getCreators().add(creator);
    }
}

